package nz.ac.massey.javaecs;
/**
 * Entity Manager
 * Controls adding and removing entities, and registration and unregistration of components to specific entities.
 * 
 * Contributors:
 * Brychan Dempsey - brychand@hotmail.com
 * 
 */

import java.util.BitSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.ArrayList;

// Define the manager classes internally - should be moved to seperate source files as appropriate
/**
 * Manages data from the perspective of the entity.
 * <p>
 * I.e. Controls adding and removing entities, and registration and deregistration of components.
 */
class EntityManager{
    private Queue<Entity> unusedEntities;
    private List<BitSet> entityRegistrations;
    private int maxSize = 1024;

    /**
     * Initialise the EntityManager with the default max size of 1024
     */
    public EntityManager(){
        unusedEntities = new LinkedList<>();
        entityRegistrations = new ArrayList<>();

        for (int i = 0; i < maxSize; i++) {
            unusedEntities.add(new Entity(i));
            entityRegistrations.add(null); // Leave bitsets out as if they are null the entity isn't initialised properly
        }
    }

    /***
     * Initialise the EntityManager with the provided maximum size
     * @param maxEntities the maximum number of entities to allow
     */
    public EntityManager(int maxEntities){
        maxSize = maxEntities;
        unusedEntities = new LinkedList<>();
        entityRegistrations = new ArrayList<>();
        
        for (int i = 0; i < maxEntities; i++) {
            unusedEntities.add(new Entity(i));
            entityRegistrations.add(null);
        }
    }

    /**
     * Creates a new entity
     * @return the index of the new entity, or -1 if there is no more available entities
     */
    protected Entity addEntity(){
        if (unusedEntities.size() == 0){
            Engine.writeErr("No available space to create a new entity");
            return null;
        }
        Entity result = unusedEntities.remove();
        entityRegistrations.set(result.getValue(), new BitSet());
        return result;
    }

    /**
     * Gets the current maximum size
     * @return the value of currentSize
     */
    protected Integer getMaxSize(){
        return maxSize;
    }

    /**
     * Gets the BitSet containing the registrations of the entity
     * @param entity the entity whose BitSet to retrieve
     * @return the BitSet of the provided entity
     */
    protected BitSet getRegistrations(Entity entity){
        try{
            return entityRegistrations.get(entity.getValue());
        }
        catch (IndexOutOfBoundsException e){
            Engine.writeErr("Index out of bounds error getting registrations for " + entity.getValue() + ";\nThe entity might not exist");
            return new BitSet(); // Using a blank BitSet will retain data safety (that is, no data will be modified)
        }
    }

    /**
     * Registers the specified component index to the entity
     * @param component the index of the component to register
     * @param entity the entity to register
     * @return true if the operation was successful
     */
    protected boolean registerComponent(int component, Entity entity){
        if (entity.getValue() >= maxSize){
            Engine.writeErr("Attempted to assign a component to non-existent entity: " + entity.getValue());
            return false;
        }
        else if (entityRegistrations.get(entity.getValue()).get(component))
        {
            Engine.writeErr("Entity is already assigned to the component");
            return false;
        }
        else{
            entityRegistrations.get(entity.getValue()).set(component);
            return true;
        }
    }

    /**
     * Adds the entity index back into unusedEntities, and sets the registrations to null
     * <p>
     * <b>Does not handle associated data</b> Use the method in ECS to remove entities cleanly
     * @param entity the entity to remove
     */
    protected void removeEntity(Entity entity){
        unusedEntities.add(entity);
        entityRegistrations.set(entity.getValue(), null);
    }

    /**
     * Sets the entity's registrations to the provided BitSet
     * @param entity the entity to set
     * @param registrations the preset registrations
     */
    protected void setRegistrations(Entity entity, BitSet registrations){
        entityRegistrations.set(entity.getValue(), registrations);
    }

    /**
     * Unregisters the specified component from the entity
     * <p>
     * <b>Does not handle component data</b> Use the method in ECS to remove components cleanly
     * @param component the component index to remove
     * @param entity the entity to remove
     * @return true if successful
     */
    protected boolean unregisterComponent(int component, Entity entity){
        try{
            entityRegistrations.get(entity.getValue()).clear(component);
            return true;
        }
        catch (NullPointerException e){
            Engine.writeErr("Entity not initialised");
            return false;
        }
        catch (IndexOutOfBoundsException e)
        {
            return false;
        }
    }

    /**
     * Resizes the currentSize of the entity manager.
     * @param newSize the new maximum size
     * @param systemManager reference to the instanced SystemManager
     * @param componentManager reference to the insanced ComponentManager
     * @return true if the operation succeeded, otherwise false
     */
    protected boolean resize(int newSize, SystemManager systemManager, ComponentManager componentManager){
        if (newSize < maxSize - unusedEntities.size()){
            Engine.writeErr("Attempted to resize the maximum entity count to a number smaller than the current assigned entity count.");
            return false;
        }
        else if (newSize == maxSize){
            Engine.writeErr("Attempted to set the newSize to the current size");
            return true;
        }
        else{
            // Consistency should be maintained.
            // This is computationally expensive; we must re-order every assigned entity above newSize, if the newSize is smaller
            if (newSize < maxSize){
                List<Entity> outOfBounds = new ArrayList<>();
                for (int i = newSize; i < maxSize; i++) {
                    Entity entityI = Entity.asEntity(i);
                    if (!unusedEntities.remove(entityI)){
                        // Could not remove element, as it didn't exist (already assigned).
                        // must find it and reassign it a new in-bounds value
                        outOfBounds.add(entityI);
                    }
                }
                for (Entity integer : outOfBounds) {
                    Entity newPos = addEntity();
                    setRegistrations(newPos, getRegistrations(integer));
                    // Invoke an update in the SystemManager
                    systemManager.entityDestroyed(integer);
                    systemManager.entityRegistrationsChanged(newPos, getRegistrations(newPos));
                    // Invoke the change in the components
                    componentManager.moveAllComponentData(integer, newPos, getRegistrations(integer));
                }
                for (int i = newSize; i < maxSize; i++) {
                    // Remove out-of-bounds data
                    entityRegistrations.remove(newSize);
                }
            }
            else{
                // Init unassigned values
                for (int i = maxSize; i < newSize; i++) {
                    unusedEntities.add(new Entity(i));
                    entityRegistrations.add(new BitSet());
                }
            }
            // Finally, set the current size
            maxSize = newSize;
            return true;
        }
    }
}